
/*
 * Guts of the paging system for ftp paging. 
 * Bashuman Deb : bdev at hss.hns.com
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

#include "curl_adap_paging.h"

/*
 * All the static function prototypes. I somehow always end up with one of
 * these humungous list when I am coding :(
 */

static int
                curl_adap_write_callback(void *, size_t, size_t, void *);
static int
                curl_adap_write_nothing(void *, size_t, size_t, void *);
static int
                setup_fresh_curl(struct ftp_stream *);
static int 
curl_adap_file_length_recorder(void *clientp, double dltotal, double dlnow, double ultotal, double ulnow);
static struct curl_paging_buff_array *
                curl_adap_paging_get(struct ftp_stream * the_stream, size_t size);
void static
                curl_paging_free_paging_encap(struct curl_paging_encap * to_free);
static void
                curl_paging_free_page_list(struct curl_page_elem * head);
static void
                curl_paging_free_paging_encap(struct curl_paging_encap * to_free);
static struct curl_paging_buff_array *
                curl_paging_get_segment_from_page(struct curl_page_elem * the_page, unsigned int *start, unsigned int *end, struct curl_paging_buff_array * head);
static void
                curl_paging_free_buff_array(struct curl_paging_buff_array * head);
struct curl_paging_buff_array *
                curl_paging_get_range(struct ftp_stream * the_stream, unsigned int start_byte, unsigned int end_byte);
static struct curl_paging_buff_array *
                curl_paging_get_tail_from_page(struct curl_page_elem * the_page, unsigned int *start, unsigned int *end, struct curl_paging_buff_array * head);
static struct curl_paging_buff_array *
                curl_paging_get_head_from_page(struct curl_page_elem * the_page, unsigned int *start, unsigned int *end, struct curl_paging_buff_array * head);
static struct curl_paging_buff_array *
                curl_paging_enter_into_list(struct curl_paging_buff_array * head, struct curl_paging_buff_array * to_insert);
static int
                curl_paging_download_range(struct ftp_stream * the_stream, unsigned int start_byte, unsigned int end_byte);
static int
curl_paging_no_overlap(struct curl_paging_encap * the_paging_encap, unsigned int start_byte, unsigned int end_byte, unsigned int closest_to_head, unsigned int closest_to_tail);
static int      curl_paging_head_overlap(struct curl_page_elem * overlap_page, unsigned int start_byte, unsigned int end_byte, unsigned int closest_to_head, unsigned int closest_to_tail);
static int 
curl_paging_tail_overlap(struct curl_page_elem * overlap_page, unsigned int start_byte, unsigned int end_byte, unsigned int closest_to_head, unsigned int closest_to_tail);
static int      curl_paging_multi_overlap(struct curl_page_elem * tail_overlap, struct curl_page_elem * head_overlap, unsigned int start_byte, unsigned int end_byte, unsigned int closest_to_head, unsigned int closest_to_tail);
static struct curl_page_elem *
curl_paging_allocate_new_page(struct curl_paging_encap * the_paging_encap, unsigned int length);
static void 
curl_paging_free_page(struct curl_page_elem * to_free);
static struct curl_page_elem *
curl_paging_find_lru(struct curl_paging_encap * the_paging_encap);
static int      curl_paging_load_in_page(struct curl_page_elem * page, unsigned int start, unsigned int end);
static int 
curl_paging_insert_page(struct curl_paging_encap * the_paging_encap, struct curl_page_elem * to_insert);


void 
do_curl_paging_banner()
{
	printf("FTP Paging support for miniunz Version. 0.01 written by Bashuman Deb\n");
	printf("See zlib license for usage details\n");
}

/*
 * fopen callback. Initializes the curl, and basic data structures and
 * returns the new ftp stream structure
 */
struct ftp_stream *
curl_adap_open_ftp_url(void *opaque, const char *urlname)
{
	struct ftp_stream *to_ret;
	static int      global_curl_inited = 0;


	if (!global_curl_inited)
	{
		do_curl_paging_banner();
		curl_global_init(CURL_GLOBAL_DEFAULT);
		global_curl_inited = 1;
	}
	if (!(FTP_URL(urlname)))
	{
		return NULL;
	}
	/*
	 * it is an url name so... Assign a state struct for the ftp url
	 */
	to_ret = (struct ftp_stream *) calloc(1, sizeof(struct ftp_stream));
	if (!to_ret)
		return to_ret;

	to_ret->curl_handle_inited = 0;
	to_ret->url_name = (char *) calloc(1, strlen(urlname) + 1);
	if (to_ret->url_name == NULL)
	{
		FREE_FTP_STREAM(to_ret);
		return NULL;
	}
	strcpy(to_ret->url_name, urlname);
	/* Now setup the paging structure */
	to_ret->paging_encap = (struct curl_paging_encap *) calloc(1, sizeof(struct curl_paging_encap));
	if (to_ret->paging_encap == NULL)
	{
		FREE_FTP_STREAM(to_ret);
		return NULL;
	}
	to_ret->paging_encap->max_number_of_pages = CURL_ADAP_MAX_PAGES;
	to_ret->paging_encap->max_individual_page_size = CURL_ADAP_PAGE_SIZE;
	to_ret->paging_encap->mummy = to_ret;	/* backwards reference */
	if (setup_fresh_curl(to_ret) < 0)
	{
		FREE_FTP_STREAM(to_ret);
		return NULL;
	}
	return to_ret;
}


static void
curl_paging_free_paging_encap(struct curl_paging_encap * to_free)
{
	if (to_free->page_head)
	{
		curl_paging_free_page_list(to_free->page_head);
	}
	free(to_free);
	return;
}

static void
curl_paging_free_page_list(struct curl_page_elem * head)
{
	if (head->next)
	{
		curl_paging_free_page_list(head->next);
	}
	free(head->buffer);
	free(head);
}


curl_adap_read(void *stream, void *buf, unsigned int size)
{

	struct ftp_stream *the_stream = (struct ftp_stream *) stream;
	CURLcode        res;
	unsigned int    start_byte, end_byte;
	unsigned int    saved_copied_bytes, bytes_to_copy;
	unsigned char  *returned_data;
	struct curl_paging_buff_array *head, *iterator;
	int             ret;
	int             copied_counter = 0;

	if (the_stream->curl_handle_inited == 0)
	{
		return -1;
	}
	head = curl_adap_paging_get(the_stream, size);
	if (NULL == head)
	{
		return -1;
	}
	for (iterator = head; iterator; iterator = iterator->next)
	{
		if ((copied_counter + iterator->buffer_size) > size)
		{
			bytes_to_copy = size - copied_counter;
		} else
		{
			bytes_to_copy = iterator->buffer_size;
		}

		memcpy((unsigned char *) buf + copied_counter, iterator->buffer, bytes_to_copy);
		copied_counter += bytes_to_copy;
		if (copied_counter > size)
			iterator = NULL;
	}

	/* increment the pointer */
	the_stream->current_pointer += size;
	curl_paging_free_buff_array(head);
	return size;
}

static void
curl_paging_free_buff_array(struct curl_paging_buff_array * head)
{
	if (head->next)
	{
		curl_paging_free_buff_array(head->next);
	}
	free(head);
}



/*
 * Free the associated data structures associated with the stream
 */
int
curl_adap_close_ftp(void *stream)
{
	struct ftp_stream *the_stream = (struct ftp_stream *) stream;
	/*
	 * first release the data structures within the stream
	 */
	if (the_stream->curl_handle_inited)
	{
		curl_easy_cleanup(the_stream->the_present_curl);
		the_stream->curl_handle_inited = 0;
	}
	FREE_FTP_STREAM(the_stream);
	return 0;
}

static int
curl_adap_write_nothing(void *buffer, size_t size, size_t nmemb, void *stream)
{
	return size * nmemb;
}

/*
 * Bad bad bad kludge callback to record the length of the file in question.
 * We need the length to seek across the file
 */
static int
curl_adap_file_length_recorder(void *clientp, double dltotal, double dlnow, double ultotal, double ulnow)
{

	struct ftp_stream *the_stream = (struct ftp_stream *) clientp;
	/* We have the length of the file */
	the_stream->file_length = (unsigned int) dltotal;
	/* We want to stop the transfer */
	return 1;
}

/*
 * this is used to record the data when curl is handing it to us
 */
static int
curl_adap_write_callback(void *buffer, size_t size, size_t nmemb, void *stream)
{

	int             i;
	struct ftp_stream *the_stream = (struct ftp_stream *) stream;
	int             size_to_copy;
	/*
	 * take the elements and iterate through them and write them to the
	 * buffer
	 */
	if (size * nmemb > (unsigned int) the_stream->buffer_size)
		size_to_copy = the_stream->buffer_size;
	else
		size_to_copy = size * nmemb;

	memcpy(the_stream->buffer + the_stream->copied_bytes, buffer, size_to_copy);
	the_stream->copied_bytes += size_to_copy;


	return size * nmemb;
}

/*
 * Curl setup function. Sets up a new curl and finds out the length of the
 * file in question
 */
static int
setup_fresh_curl(struct ftp_stream * the_stream)
{
	CURLcode        res;

	if (the_stream->curl_handle_inited)
	{
		curl_easy_cleanup(the_stream->the_present_curl);
		the_stream->curl_handle_inited = 0;
	}
	the_stream->the_present_curl = curl_easy_init();
	if (!the_stream->the_present_curl)
	{
		/* We have failed you master ! */
		return -1;
	}
#ifdef DEBUG_NET
	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_VERBOSE, TRUE);
#endif

	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_NOPROGRESS, FALSE);
	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_URL, the_stream->url_name);
	/* Now set up the file length */

	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_FILE, the_stream);
	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_WRITEFUNCTION, curl_adap_write_nothing);
	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_PROGRESSFUNCTION, curl_adap_file_length_recorder);
	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_PROGRESSDATA, the_stream);
	res = curl_easy_perform(the_stream->the_present_curl);



	/*
	 * We assume that the length of the file has been set and the
	 * transfer aborted
	 */
	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_WRITEFUNCTION, curl_adap_write_callback);
	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_FILE, the_stream);
	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_NOPROGRESS, 1);

	curl_easy_setopt(the_stream->the_present_curl, CURLOPT_FRESH_CONNECT, 1);


	the_stream->curl_handle_inited = 1;

	return 0;
}

/* Seek.... */
int
curl_adap_seek_end(struct ftp_stream * the_stream, unsigned long offset)
{
	the_stream->current_pointer = ((unsigned long) (the_stream->file_length) - offset);
	return 0;

}




/*
 * fread callback. Tries to retrieve data from page list. If failure,
 * requests a download,and then retries. If still fail, we are hosed
 */
static struct curl_paging_buff_array *
curl_adap_paging_get(struct ftp_stream * the_stream, size_t size)
{

	unsigned int    start_byte;
	unsigned int    end_byte;
	int             ret;
	struct curl_paging_buff_array *to_ret;
	/* Find out the byte offsets required */
	start_byte = (unsigned int) (the_stream->current_pointer);
	end_byte = (unsigned int) (the_stream->current_pointer + size);

	/* First find out whether the range we want is available or not */
	to_ret = curl_paging_get_range(the_stream, start_byte, end_byte);
	if (NULL == to_ret)
	{
		/* Page miss */
		ret = curl_paging_download_range(the_stream, start_byte, end_byte);
		if (ret <= 0)
		{
			return NULL;
		}
		/* Get the stuff into the pages */
		to_ret = curl_paging_get_range(the_stream, start_byte, end_byte);
	}
	return to_ret;
}


/*
 * Main get function called by get. Iterates through all pages and creates a
 * linked list of buffers to be used
 */
struct curl_paging_buff_array *
curl_paging_get_range(struct ftp_stream * the_stream, unsigned int start_byte, unsigned int end_byte)
{
	struct curl_paging_encap *the_page_structure;
	struct curl_page_elem *iterator;
	struct curl_paging_buff_array *to_ret = NULL, **element;
	unsigned int    remaining_start, remaining_end;
	int             have_head, have_tail;

	the_page_structure = the_stream->paging_encap;

	/*
	 * First find if the range is available in the whole list of files
	 */
	remaining_start = start_byte;
	remaining_end = end_byte;
	FOR_EACH_PAGE_STREAM(the_page_structure, iterator)
	{
		/*
		 * Either the whole range will be covered in the page or we
		 * will have a head or a tail
		 */
		have_head = have_tail = 0;
		if ((remaining_start >= iterator->begin) && (remaining_start <= iterator->end) && (remaining_end > iterator->end))
		{
			have_head = 1;
		}
		if ((remaining_end <= iterator->end) && (remaining_end >= iterator->begin) && (remaining_start < iterator->begin))
		{
			have_tail = 1;
		}
		if ((remaining_start >= iterator->begin) && (remaining_end <= iterator->end))
		{
			have_head = have_tail = 1;
		}
		if (have_head && have_tail)
		{
			to_ret = curl_paging_get_segment_from_page(iterator, &remaining_start, &remaining_end, to_ret);
		} else if (have_tail)
		{
			to_ret = curl_paging_get_tail_from_page(iterator, &remaining_start, &remaining_end, to_ret);
		} else if (have_head)
		{
			to_ret = curl_paging_get_head_from_page(iterator, &remaining_start, &remaining_end, to_ret);
		}
		if (remaining_start == remaining_end)
		{
			break;
		}
	}
	if ((remaining_start != remaining_end))
	{
		if (to_ret)
			curl_paging_free_buff_array(to_ret);
		to_ret = NULL;

	}
	return to_ret;
}


static struct curl_paging_buff_array *
curl_paging_get_segment_from_page(struct curl_page_elem * the_page, unsigned int *start, unsigned int *end, struct curl_paging_buff_array * head)
{
	unsigned int    offset_of_start;
	struct curl_paging_buff_array *to_insert;
	to_insert = (struct curl_paging_buff_array *) calloc(1, sizeof(struct curl_paging_buff_array));
	if (NULL == to_insert)
	{
		return NULL;
	}
	offset_of_start = (unsigned int) (*start - the_page->begin);

	to_insert->offset_into_file = the_page->begin + offset_of_start;
	to_insert->buffer_size = (unsigned int) (*end - *start);
	to_insert->buffer = the_page->buffer + offset_of_start;

	/* the start and end have come to the unite */
	*start = *end;
	/* update the time on the page */
	gettimeofday(&(the_page->last_used), NULL);
	return curl_paging_enter_into_list(head, to_insert);

}

static struct curl_paging_buff_array *
curl_paging_get_tail_from_page(struct curl_page_elem * the_page, unsigned int *start, unsigned int *end, struct curl_paging_buff_array * head)
{
	struct curl_paging_buff_array *to_insert;

	to_insert = (struct curl_paging_buff_array *) calloc(1, sizeof(struct curl_paging_buff_array));
	if (NULL == to_insert)
	{
		return to_insert;
	}
	to_insert->offset_into_file = the_page->begin;
	to_insert->buffer_size = (unsigned int) (*end - to_insert->offset_into_file);
	to_insert->buffer = the_page->buffer;
	*end = the_page->begin - 1;

	/* update the time on the page */
	gettimeofday(&(the_page->last_used), NULL);
	return curl_paging_enter_into_list(head, to_insert);

}

static struct curl_paging_buff_array *
curl_paging_get_head_from_page(struct curl_page_elem * the_page, unsigned int *start, unsigned int *end, struct curl_paging_buff_array * head)
{
	struct curl_paging_buff_array *to_insert;

	to_insert = (struct curl_paging_buff_array *) calloc(1, sizeof(struct curl_paging_buff_array));
	if (NULL == to_insert)
	{
		return to_insert;
	}
	to_insert->offset_into_file = *start;
	to_insert->buffer = the_page->buffer + ((unsigned int) (*start - the_page->begin));
	to_insert->buffer_size = the_page->buffer_size - ((unsigned int) (*start - the_page->begin));
	*start = the_page->end + 1;

	/* update the time on the page */
	gettimeofday(&(the_page->last_used), NULL);
	return curl_paging_enter_into_list(head, to_insert);
}

static struct curl_paging_buff_array *
curl_paging_enter_into_list(struct curl_paging_buff_array * head, struct curl_paging_buff_array * to_insert)
{
	struct curl_paging_buff_array *iterator, *tail;
	int             done = 0;
	if (head == NULL)
	{
		return to_insert;
	}
	for (iterator = head; iterator; iterator = iterator->next)
	{
		tail = iterator;
		if (iterator->offset_into_file > to_insert->offset_into_file)
		{
			/*
			 * We have found the element whose prev link this is
			 * meant to be
			 */
			iterator->prev->next = to_insert;
			to_insert->prev = iterator->prev;
			to_insert->next = iterator;
			iterator->prev = to_insert;
			return head;
		}
	}

	/* the element is to be inserted int the tail */
	tail->next = to_insert;
	to_insert->next = NULL;
	return head;
}

static int
curl_paging_download_range(struct ftp_stream * the_stream, unsigned int start_byte, unsigned int end_byte)
{
	struct curl_paging_encap *the_paging_encap = the_stream->paging_encap;
	unsigned int    length_do_download;
	unsigned int    closest_upper, closest_lower;
	struct curl_page_elem *head_overlap, *tail_overlap, *iterator;
	int             have_head, have_tail;
	int             ret;
	unsigned int    closest_to_head, closest_to_tail;

	closest_to_head = closest_to_tail = UINT_MAX;

	length_do_download = end_byte - start_byte;
	have_head = have_tail = (int) head_overlap = tail_overlap = NULL;

	if (length_do_download > (the_paging_encap->max_individual_page_size * the_paging_encap->max_number_of_pages))
	{
		fprintf(stderr, "Too large a download\n");
		return -1;
	}
	/* Now check the stuff */
	FOR_EACH_PAGE_STREAM(the_paging_encap, iterator)
	{
		have_head = have_tail = 0;
		if ((start_byte >= iterator->begin) && (start_byte <= iterator->end) && (end_byte > iterator->end))
		{
			have_head = 1;
		}
		if ((end_byte < iterator->end) && (end_byte >= iterator->begin) && (start_byte < iterator->begin))
		{
			have_tail = 1;
		}
		if (have_tail)
		{
			tail_overlap = iterator;
		} else if (have_head)
		{
			head_overlap = iterator;
		}
		/*
		 * Determine the closest neighbour we have also
		 */
		if (!(have_head || have_tail))
		{
			if (iterator->begin > end_byte)
			{
				if (iterator->begin < closest_to_tail)
					closest_to_tail = iterator->begin;
			}
			if (iterator->end < start_byte)
			{
				if (iterator->end < closest_to_head)
				{
					closest_to_head = iterator->end;
				}
			}
		}
	}

	if (tail_overlap && head_overlap)
	{
		/*
		 * We check out in which page the larger portion is and free
		 * that page and include the demanded portion in a new page
		 */
		ret = curl_paging_multi_overlap(tail_overlap, head_overlap, start_byte, end_byte, closest_to_head, closest_to_tail);
	} else if (head_overlap)
	{
		/*
		 * We simply download the rest of the portion in a new page.
		 * If it reaches the end of the file, or more, we delete the
		 * head overlap and allocate a page till the end of the file
		 */
		ret = curl_paging_head_overlap(head_overlap, start_byte, end_byte, closest_to_head, closest_to_tail);
	} else if (tail_overlap)
	{
		/*
		 * download from the tail upwards
		 */
		ret = curl_paging_tail_overlap(tail_overlap, start_byte, end_byte, closest_to_head, closest_to_tail);
	} else
	{
		/*
		 * simply download an eventual buffer
		 */
		ret = curl_paging_no_overlap(the_paging_encap, start_byte, end_byte, closest_to_head, closest_to_tail);
	}
	return ret;

}

static int 
curl_paging_multi_overlap(struct curl_page_elem * tail_overlap, struct curl_page_elem * head_overlap, unsigned int start_byte, unsigned int end_byte, unsigned int closest_to_head, unsigned int closest_to_tail)
{
	unsigned int    new_end;
	/* Get rid of the head overlap page */
	curl_paging_free_page(head_overlap);
	new_end = tail_overlap->end - 1;
	return curl_paging_no_overlap(tail_overlap->mummy, start_byte, new_end, closest_to_head, closest_to_tail);

}

static int 
curl_paging_tail_overlap(struct curl_page_elem * overlap_page, unsigned int start_byte, unsigned int end_byte, unsigned int closest_to_head, unsigned int closest_to_tail)
{
	unsigned int    new_end_of_page = overlap_page->end;
	struct curl_paging_encap *the_paging_encap = overlap_page->mummy;
	/* Free the overlap page buffer, allocate a new buffer */
	curl_paging_free_page(overlap_page);
	return curl_paging_no_overlap(the_paging_encap, start_byte, new_end_of_page, closest_to_head, closest_to_tail);
}

static int 
curl_paging_head_overlap(struct curl_page_elem * overlap_page, unsigned int start_byte, unsigned int end_byte, unsigned int closest_to_head, unsigned int closest_to_tail)
{

	unsigned int    new_start_byte;
	unsigned int    delta;
	struct curl_paging_encap *the_paging_encap = overlap_page->mummy;
	/*
	 * Check whether the delta addition to the page is more that 10% of
	 * page size. If it is less, download after the overlap page
	 */
	delta = end_byte - overlap_page->end;
	if (delta < (the_paging_encap->max_individual_page_size / 10))
	{
		new_start_byte = overlap_page->end + 1;
		if (new_start_byte >= end_byte)
			end_byte++;
	} else
	{
		new_start_byte = overlap_page->begin;
		/* Free the overlap page buffer, allocate a new buffer */
		curl_paging_free_page(overlap_page);
	}
	return curl_paging_no_overlap(the_paging_encap, new_start_byte, end_byte, closest_to_head, closest_to_tail);
}

static int
curl_paging_no_overlap(struct curl_paging_encap * the_paging_encap, unsigned int start_byte, unsigned int end_byte, unsigned int closest_to_head, unsigned int closest_to_tail)
{
	unsigned int    the_size_of_segment;
	struct curl_page_elem *the_new_page;
	unsigned int    start_to_download, end_to_download;
	int             ret;

	the_size_of_segment = end_byte - start_byte;
	if (the_size_of_segment > (the_paging_encap->max_individual_page_size) * (the_paging_encap->max_number_of_pages))
	{
		fprintf(stderr, "Desired area bigger that allowed memory (%lu)\n", the_size_of_segment);
		/* Sorry. We are closed down here */
		return -1;
	}
	/* Now we actually download the stuff */

	if (the_size_of_segment <= (unsigned int) (the_paging_encap->max_individual_page_size))
	{
		/*
	 	* check size of file. If less than size per page get the whole file.
	       	*/
		if ((unsigned int) (the_paging_encap->max_individual_page_size) > the_paging_encap->mummy->file_length)
		{
			/* Just download the whole file */
			the_new_page = curl_paging_allocate_new_page(the_paging_encap, the_paging_encap->mummy->file_length);
			if (NULL == the_new_page)
			{
				return -1;
			}
			start_to_download = 0.0;
			end_to_download = 0.0;
		} else
		{
			/*
		  	* We will adjust the size of the stuff to download
		  	* The approach will be that we append more bytes to the end of the
			 * tail
		  	*/
			start_to_download = start_byte;
			end_to_download = start_byte + (unsigned int) the_paging_encap->max_individual_page_size - 1;
			if (end_to_download > closest_to_tail)
				end_to_download = closest_to_tail - 1;

			if (end_to_download > the_paging_encap->mummy->file_length)
				end_to_download = the_paging_encap->mummy->file_length;

			the_new_page = curl_paging_allocate_new_page(the_paging_encap, end_to_download - start_byte + 1);
			if (NULL == the_new_page)
			{
				return -1;
			}
		}

	} else
	{
		/* We will download the complete thing in one buffer */
		the_new_page = curl_paging_allocate_new_page(the_paging_encap, the_size_of_segment);
		if (NULL == the_new_page)
		{
			return -1;
		}
		start_to_download = start_byte;
		end_to_download = end_byte;
	}

	ret = curl_paging_load_in_page(the_new_page, start_to_download, end_to_download);
	if (ret <= 0)
	{
		curl_paging_free_page(the_new_page);
	}
	return ret;


}

static struct curl_page_elem *
curl_paging_allocate_new_page(struct curl_paging_encap * the_paging_encap, unsigned int length)
{
	struct curl_page_elem *lru_page, *to_ret;
	if (the_paging_encap->num_of_allocated_bytes > (the_paging_encap->max_individual_page_size * the_paging_encap->max_number_of_pages))
	{
		/*
		 * We are all booked dudes.. So we free the lru page
		 */
		lru_page = curl_paging_find_lru(the_paging_encap);
		if (lru_page != NULL)
			curl_paging_free_page(lru_page);

	}
	to_ret = (struct curl_page_elem *) calloc(1, sizeof(struct curl_page_elem));
	if (NULL == to_ret)
		return to_ret;

	to_ret->buffer = (unsigned char *) calloc(1, (unsigned int) length);
	if (NULL == to_ret->buffer)
	{
		free(to_ret);
		return NULL;
	}
	to_ret->buffer_size = (unsigned int) length;
	to_ret->mummy = the_paging_encap;
	gettimeofday(&(to_ret->last_used), NULL);
	the_paging_encap->num_of_allocated_bytes += (unsigned int) length;

	return to_ret;

}

static void 
curl_paging_free_page(struct curl_page_elem * to_free)
{
	struct curl_paging_encap *the_paging_encap = to_free->mummy;

	if (to_free == the_paging_encap->page_head)
	{
		the_paging_encap->page_head = the_paging_encap->page_head->next;
	}
	free(to_free->buffer);
	the_paging_encap->num_of_allocated_bytes -= to_free->buffer_size;
	if (to_free->prev)
		to_free->prev->next = to_free->next;
	if (to_free->next)
		to_free->next->prev = to_free->prev;
	free(to_free);
}

static struct curl_page_elem *
curl_paging_find_lru(struct curl_paging_encap * the_paging_encap)
{
	struct curl_page_elem *iterator, *least_used = NULL;

	FOR_EACH_PAGE_STREAM(the_paging_encap, iterator)
	{
		if (NULL == least_used)
		{
			least_used = iterator;
		} else
		{
			if (iterator->last_used.tv_sec < least_used->last_used.tv_sec)
			{
				least_used = iterator;
			} else if (iterator->last_used.tv_sec == least_used->last_used.tv_sec)
			{
				if (iterator->last_used.tv_usec < least_used->last_used.tv_usec)
				{
					least_used = iterator;
				}
			}
		}

	}
	return least_used;
}

static int 
curl_paging_load_in_page(struct curl_page_elem * page, unsigned int start, unsigned int end)
{
	/*
	 * Check if the curl is already initialized, if not close the thing,
	 * get a new curl start reading
	 */
	CURLcode        res;
	int             ret;
	char            range[500];
	unsigned int    start_byte, end_byte;
	unsigned int    saved_copied_bytes;
	struct ftp_stream *the_stream;

	the_stream = page->mummy->mummy;

	/* Open up the curl */
	if (the_stream->curl_handle_inited == 0)
	{
		return -1;
	}
	/* Now we start up the actual range definition */
	if ((0 == start) && (0 == end))
	{
		page->begin = 0;
		page->end = the_stream->file_length;
	} else
	{
		start_byte = start;
		end_byte = end;
		page->begin = start;
		page->end = end;
		sprintf(range, "%d-%d", start_byte, end_byte);
		curl_easy_setopt(the_stream->the_present_curl, CURLOPT_RANGE, range);
	}

	the_stream->buffer = (unsigned char *) page->buffer;
	the_stream->buffer_size = page->buffer_size;
	the_stream->copied_bytes = 0;

	/* Now start the transfer */
	res = curl_easy_perform(the_stream->the_present_curl);
	if (res != CURLE_OK)
	{
#ifdef DEBUG_NET
		fprintf(stderr, "Curl error %d\n", res);
#endif
	}
	/* We assume the the amount of data required has come in */
	if (the_stream->copied_bytes <= 0)
	{
		fprintf(stderr, "Read error %d\n", res);
		return -1;
	}
	saved_copied_bytes = the_stream->copied_bytes;
	the_stream->copied_bytes = 0;

	/*
	 * Now insert the page into the page chain
	 */
	the_stream->buffer = NULL;
	the_stream->buffer_size = 0;

	ret = curl_paging_insert_page(the_stream->paging_encap, page);
	if (ret != 0)
	{
		return -1;
	}
	return saved_copied_bytes;;
}

static int 
curl_paging_insert_page(struct curl_paging_encap * the_paging_encap, struct curl_page_elem * to_insert)
{

	struct curl_page_elem *iterator, *tail;
	if (the_paging_encap->page_head)
	{
		FOR_EACH_PAGE_STREAM(the_paging_encap, iterator)
		{
			tail = iterator;
			if (iterator->begin > to_insert->begin)
			{
				if (iterator->prev)
					iterator->prev->next = to_insert;
				else
				{
					/*
					 * this is to become the head of the
					 * list
					 */
					the_paging_encap->page_head = to_insert;
				}
				to_insert->prev = iterator->prev;
				to_insert->next = iterator;
				iterator->prev = to_insert;
				return 0;
			}
		}
		tail->next = to_insert;
		to_insert->prev = tail;
		to_insert->next = NULL;
		return 0;
	} else
	{
		the_paging_encap->page_head = to_insert;
		to_insert->prev = NULL;
		to_insert->next = NULL;
		return 0;
	}

	return 0;
}
