
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>


#ifndef CURL_ADAP_PAGING_H
#define CURL_ADAP_PAGING_H	1


/*
 * The next two macros actually determine as to how much memory the
 * application is going to consume at runtime
 */
#ifndef CURL_ADAP_MAX_PAGES
#define CURL_ADAP_MAX_PAGES	5	/* Number of pages to use */
#endif

#ifndef CURL_ADAP_PAGE_SIZE
#define CURL_ADAP_PAGE_SIZE	1024*1024	/* How much memory to use for
						 * per page */
#endif

#ifndef CURL_ADAP_MINIMUM_TRANSFER
#define CURL_ADAP_MINIMUM_TRANSFER	100*1024
#endif

#define FTP_URL(x)	((x[0] == 'f') && \
		        (x[1] == 't') && \
	       		(x[2] == 'p') && \
	 		(x[3] == ':') && \
			(x[4] == '/') && \
			(x[5] == '/'))


#define FREE_FTP_STREAM(x)	do { \
					if (x) \
					{ \
						if(x->url_name) \
							free(x->url_name) ; \
						if (x->paging_encap) \
						{ \
							curl_paging_free_paging_encap(x->paging_encap) ; \
						} \
						free(x) ;  \
						x = NULL ; \
					} \
				} while(0)




#define FOR_EACH_PAGE(x,y)	for(y=x;y;y=y->next)
#define FOR_EACH_IN_BUFFER_ARRAY(x,y) for(y=x;y;y=y->next)
#define FOR_EACH_PAGE_STREAM(x,y)	for(y=(x)->page_head;y;y=y->next)

/*
 * Structure defining a single page struture member names are self
 * explanatory.
 */
struct curl_page_elem
{
	struct curl_page_elem *prev;
	struct curl_page_elem *next;
	unsigned char  *buffer;
	unsigned int    buffer_size;
	unsigned int    begin;	/* Offset in complete file */
	unsigned int    end;	/* End in complete file */
	struct timeval  last_used;
	struct curl_paging_encap *mummy;	/* Back reference */

};

/*
 * The whole paging encapsulation for one particular file
 */
struct curl_paging_encap
{
	unsigned int    max_number_of_pages;
	unsigned int    max_individual_page_size;
	unsigned int    num_of_allocated_bytes;

	struct curl_page_elem *page_head;	/* doubly linked list to all
						 * pages */
	struct ftp_stream *mummy;
};


/*
 * a single element of a linked list which is yielded when we search for
 * contigious data across a bunch of pages
 */

struct curl_paging_buff_array
{
	struct curl_paging_buff_array *prev;
	struct curl_paging_buff_array *next;
	unsigned int    offset_into_file;
	unsigned char  *buffer;
	unsigned int    buffer_size;
};






struct ftp_stream
{
	char           *url_name;	/* The ftp url in question.
					 * Dynamically allocated */
	long            current_pointer;	/* the position in the file */
	int             curl_handle_inited;	/* Whether we already have a
						 * curl here */
	CURL           *the_present_curl;	/* The curl */
	unsigned char  *buffer;	/* output buffer */
	int             buffer_size;	/* the size of the output buffer */
	int             bytes_to_be_read;	/* the number of bytes in the
						 * read request */
	int             copied_bytes;	/* number of bytes copied */
	unsigned int    file_length;	/* Length of file to be downloaded */


	/* The paging structure for the Stream */
	struct curl_paging_encap *paging_encap;

};




int
                curl_adap_read(void *stream, void *buf, unsigned int size);
struct ftp_stream *
                curl_adap_open_ftp_url(void *opaque, const char *urlname);
int
                curl_adap_close_ftp(void *stream);
int
                curl_adap_seek_end(struct ftp_stream *, unsigned long);


#endif				/* CURL_ADAP_PAGING_H */
